import os

import addict
import pytest

from act.scio.plugins import vulnerabilities


@pytest.mark.asyncio  # type: ignore
async def test_vulnerabilites() -> None:
    """test for plugins"""

    nlpdata = addict.Dict()
    nlpdata.content = """
    CVE-1991-1234 lorem ipsum
    CVE-1992-12345 lorem ipsum
    CVE-1993-123456 lorem ipsum
    CVE-1994-1234567 lorem ipsum
    CVE-1995-12345678 lorem ipsum
    CVE-1994-12 lorem ipsum
    CVE-123-1234 lorem ipsum
    MS15-132 lorem ipsum
    """

    plugin = vulnerabilities.Plugin()
    plugin.configdir = os.path.join(
        os.path.dirname(__file__), "../act/scio/etc/plugins"
    )
    res = await plugin.analyze(nlpdata)

    assert "CVE-1991-1234" in res.result.cve
    assert "CVE-1992-12345" in res.result.cve
    assert "CVE-1993-123456" in res.result.cve
    assert "CVE-1994-1234567" in res.result.cve
    assert "MS15-132" in res.result.msid
    assert "CVE-123-1234" not in res.result.cve
    assert "CVE-1994-12" not in res.result.cve
    assert "CVE-1995-12345678" not in res.result.cve



"""Plugin for finding possible threat actors based on sentences"""

from typing import List, Set, Text

import addict
import nltk
import nltk.stem

from act.scio.plugin import BasePlugin, Result


class Plugin(BasePlugin):
    """Plugin to detect threat actors"""

    name = "nlp_actors"
    info = "Extract actors based on language from a body of text"
    version = "0.1"
    dependencies: List[Text] = ["pos_tag"]

    async def analyze(self, nlpdata: addict.Dict) -> Result:

        res = addict.Dict()

        threat_stem_postfix = {
            "threat",  # threat
            "crimin",  # criminal, criminals
            "crime",  # crime
            "espionage",  # espionage
            "hack",  # hack, hacking,
            "hacker",  # hacker, hackers
            "crack",  # cracking, crack
            "cracker",  # cracker, crackers
            "adversari",  # adversary, adversaries
            "terrorist",  # terrorist, terrorists
        }

        group_stem_postfix = {
            "group",  # group, groups
            "actor",  # actor, actors
            "unit",  # unit, untis
            "agent",  # agent, agents
            "organ",  # organization, organizations
        }

        false_positive_filter = [
            "top",
            "unknown",
            "cyber",
        ]  # "top threat groups", "cyber threat actors" etc...

        possible_ta_tag_types = {"NNP", "NNPS", "NN", "NNS"}
        possible_tag_types = {"NNP", "NNPS", "NN", "NNS", "JJ", "JJS"}
        chain_tags = {",", ":", "CC"}
        lookbefore_tags: Set[Text] = set()
        lookbefore_tags.update(chain_tags)
        lookbefore_tags.update(possible_tag_types)

        ps = nltk.stem.PorterStemmer()

        first_stage_found = False

        pos_actors: List[Text] = []
        # Look through all tokens. If any token relating to a threat actors is
        # found, look-before and collect all nouns while the tokens are nouns
        # or part of a listing.
        for i, (token, tag) in enumerate(nlpdata.pos_tag.tokens):
            if first_stage_found:
                second_stage_found = bool(
                    tag in possible_tag_types and ps.stem(token) in group_stem_postfix
                )
                if not second_stage_found:
                    first_stage_found = False
                    continue

                if nlpdata.pos_tag.tokens[i - 2][1] not in possible_tag_types:
                    first_stage_found = False
                    continue
                n = i - 1
                while (
                    n > 0
                    and len(nlpdata.pos_tag.tokens[n]) == 2
                    and nlpdata.pos_tag.tokens[n][1] in lookbefore_tags
                ):
                    n -= 1

                current_actor: List[Text] = []
                for (subtoken, pos_tag) in nlpdata.pos_tag.tokens[
                    n : i - 1  # noqa: E203
                ]:
                    # check if we have reached a separator (comma, 'and' etc)
                    # if so, we need to create a result of what we have found thus
                    # far and look for more.
                    if pos_tag in chain_tags:
                        if current_actor:
                            if valid_actor(current_actor):
                                pos_actors.append(" ".join(current_actor))
                            current_actor = []
                    elif pos_tag in possible_ta_tag_types:
                        if subtoken in false_positive_filter:
                            continue
                        current_actor.append(subtoken)
                if current_actor:
                    pos_actors.append(" ".join(current_actor))

            # check wether the current tag is of a type and in the accepted list of
            # threat group postfixes.
            first_stage_found = bool(
                tag in possible_tag_types and ps.stem(token) in threat_stem_postfix
            )

        res.actors = pos_actors
        return Result(name=self.name, version=self.version, result=res)


def valid_actor(actor: List[Text]) -> bool:
    """Check that the list of elements are actualy
    valid formatted"""

    if not actor:
        return False
    for e in actor:
        # Check that all elements start with an upper case letter
        if e and not e[0].isupper():
            return False
    return True



""" test feed download """

from act.scio.feeds import extract


def test_safe_download() -> None:
    """test for safe download"""
    assert extract.replace_unsafe_filename_characters("test%.[x y z]") == "test.x_y_z"


def test_santize_filename() -> None:
    """Test for filename lengtt and sane name"""

    wrong = "/home/scio/.cache/scio-feeds/download/June_2_6122015_Consulting_Thought_Leadership_quotProactively_Engagedâ_Questions_Executives_Should_Ask_Their_Security_Teams_quotquot-Many_breaches_occur_as_a_result_of_executive_decisions_made_woutfull_knowledge_of_the_peopleprocesses_needed_to_prevent_them__-Offersspecific_questions_that_execs_should_ask_to_understand_and_prevent_abreachquot_Jim_Aldridge_Kyrk_Content_Finalized_Global________June_26122015_Consulting_Thought_Leadership_quotProactively_Engaged_âQuestions_Executives_Should_Ask_Their_Security_Teams_quot_quot-Manybreaches_occur_as_a_result_of_executive_decisions_made_wout_fullknowledge_of_the_peopleprocesses_needed_to_prevent_them__-Offersspecific_questions_that_execs_should_ask_to_understand_and_prevent_abreachquot_Jim_Aldridge_Kyrk_Content_Finalized_GlobCaching_Out_TheValue_of_Shimcache_for_Investigators.html"  # noqa E501
    correct = "/home/scio/.cache/scio-feeds/download/June_2_6122015_Consulting_Thought_Leadership_quotProactively_Engagedâ_Questions_Executives_Should_Ask_Their_Security_Teams_quotquot-Many_breaches_occur_as_a_result_of_executive_decisions_made_woutfull_knowledge_of_the_peopleprocesses_needed_to_preve.html"  # noqa: E501

    assert extract.sanitize_filename(wrong) == correct


def test_safe_filename() -> None:
    """Test that / are correctly removed"""

    filename = "Example / This is a %title%"

    path = extract.create_storage_path(filename, "html", "/", "/tmp/", "download")

    assert path == "/tmp/download/Example__This_is_a_title.html"



""" Indicators test """

import os

import addict
import pytest

from act.scio.plugins import indicators

TEST_TEXT = """
    hXXp://my.test.no/hxxp/
    md5: be5ee729563fa379e71d82d61cc3fdcf lorem ipsum
    sha256: 103cb6c404ba43527c2deac40fbe984f7d72f0b2366c0b6af01bd0b4f1a30c74 lorem ipsum
    sha1: 3c07cb361e053668b4686de6511d6a904a9c4495 lorem ipsum
    %2fchessbase.com lorem ipsum
    %2Fchessbase.com lorem ipsum
    twitter.com lorem ipsum
    %2ftwitter.com lorem ipsum
    %2Ftwitter.com lorem ipsum
    127.0.0.1 lorem ipsum
    127[.]0[.]0[.]2 lorem ipsum
    127.0.0{.}3 lorem ipsum
    127.0.0\\.4 lorem ipsum
    ftp://files.example.com lorem ipsum
    https://www.vg.no/index.html?q=news#top lorem / ipsum
    HTTP://1.2.3.4/5-index.html / lorem ipsum
    hXXp://2.3.4.5/ lorem / ipsum
    hxxps://3.4.5.6 lorem ipsum
    4.5.6.7/gurba lorem ipsum
    5.6.7.8/9 lorem ipsum
    6.7.8.9/10 lorem ipsum
    www.nytimes3xbfgragh.onion lorem ipsum
    fe80::ea39:35ff:fe12:2d71/64 lorem ipsum
    The mail address user@fastmail.fm is not real
    www.mnemonic.no
    this.ends.in.tld.no."
"""


@pytest.mark.asyncio  # type: ignore
async def test_sectors() -> None:
    """test for plugins"""

    nlpdata = addict.Dict()
    plugin = indicators.Plugin()
    plugin.configdir = os.path.join(
        os.path.dirname(__file__), "../act/scio/etc/plugins"
    )

    nlpdata.content = TEST_TEXT

    res = await plugin.analyze(nlpdata)

    # print(res.result)

    assert res.result.md5 == ["be5ee729563fa379e71d82d61cc3fdcf"]
    assert res.result.sha256 == [
        "103cb6c404ba43527c2deac40fbe984f7d72f0b2366c0b6af01bd0b4f1a30c74"
    ]
    assert res.result.sha1 == ["3c07cb361e053668b4686de6511d6a904a9c4495"]

    assert "http://my.test.no/hxxp/" in res.result.uri
    assert "HTTP://1.2.3.4/5-index.html" in res.result.uri
    assert "https://3.4.5.6" in res.result.uri
    assert "www.mnemonic.no" in res.result.fqdn
    assert "chessbase.com" in res.result.fqdn
    assert "twitter.com" in res.result.fqdn
    assert "127.0.0.1" in res.result.ipv4
    assert "127.0.0.2" in res.result.ipv4
    assert "127.0.0.3" in res.result.ipv4
    assert "127.0.0.4" in res.result.ipv4
    assert "4.5.6.7" in res.result.ipv4
    assert "5.6.7.8/9" in res.result.ipv4net
    assert "user@fastmail.fm" in res.result.email
    assert "fe80::ea39:35ff:fe12:2d71" in res.result.ipv6



"""
Vocabulary tests
"""

import os
from typing import List, Text

import addict

from act.scio.alias import parse_aliases
from act.scio.aliasregex import normalize
from act.scio.vocabulary import Vocabulary

VOCABULARY_DATADIR = os.path.join(os.path.dirname(__file__), "vocabulary")


def test_vocabulary_sector() -> None:
    """Sector tests"""

    config = addict.Dict()
    config.alias = os.path.join(VOCABULARY_DATADIR, "sector_aliases.cfg")
    config.key_mod = "stem"
    config.primary = True

    sector = Vocabulary(config)

    assert sector["defense"] == "defence"
    assert sector.get("defense", primary=False) == "defense"
    assert sector["not_exists"] is None


def normalize_ta(name: Text) -> Text:
    return normalize(
        name,
        capitalize=True,
        uppercase_abbr=["APT", "BRONZE", "IRON", "GOLD"],
    )


def test_vocabulary_threat_actor() -> None:
    """Threat actor vocabulary tests"""

    config = addict.Dict()
    config.alias = os.path.join(VOCABULARY_DATADIR, "ta_aliases.cfg")
    config.regexfromalias = True
    config.regexmanual = r"""
        \b([a-zA-Z]{4,}\s?[-_. ](?:dragon|duke|falcon|cedar|viper|panda|bear|jackal|spider|chollima|kitten|tiger))\b
        \b((?:BRONZE|IRON|GOLD)(?:\s+|[-_.]+)[A-Z]{3,})\b"""  # noqa: E501

    ta = Vocabulary(config)

    matches: List[Text] = ta.regex_search(
        "Observed threat actors apt_32, apt_28 and Crazy Kitten",
        normalize_result=normalize_ta,
    )

    assert "APT 32" in matches
    assert "APT 28" in matches
    assert "Crazy Kitten" in matches

    assert ta.get("OceanLotus Group", primary=True) == "APT32"
    assert ta.get("OCEANLOTUS GROUP", key_mod="none") is None
    assert ta.get("oceanLotusGroup", key_mod="norm") == "OceanLotus Group"
    assert ta.get("aPT32", key_mod="lower") == "APT32"
    assert ta.get("OCEANLOTUS GROUP", primary=True) == "APT32"
    assert ta["OCEANLOTUS GROUP"] == "OceanLotus Group"
    assert ta["not_exists"] is None


def test_vocabulary_tool() -> None:
    """Tool vocabulary test"""

    config = addict.Dict()
    config.alias = os.path.join(VOCABULARY_DATADIR, "tool_aliases.cfg")

    tool = Vocabulary(config)

    assert tool.get("backdoor:java/adwind", primary=True) == "jrat"
    assert tool["backdoor:java/adwind"] == "backdoor:java/adwind"


def test_vocabulary_aliases() -> None:
    """Parse alias config test"""

    alias_line1 = "my\\:tool: "
    alias_line2 = "thetool: backdoor\\:java/adwind,comma\\,tool"

    assert parse_aliases(alias_line1)[0] == "my:tool"
    assert parse_aliases(alias_line2)[0] == "thetool"
    assert "backdoor:java/adwind" in parse_aliases(alias_line2)[1]
    assert "comma,tool" in parse_aliases(alias_line2)[1]



""" test feed download """

import io
import os

import pytest

# Unfortunately not exported
from _pytest.monkeypatch import MonkeyPatch

from act.scio import analyze, plugin


@pytest.mark.asyncio  # type: ignore
async def test_plugins(monkeypatch: MonkeyPatch) -> None:
    """test for plugins"""

    plugin_dir = os.path.join(
        os.path.dirname(os.path.realpath(__file__)), "plugins_deps"
    )

    plugins = plugin.load_external_plugins(plugin_dir)

    monkeypatch.setattr(
        "sys.stdin", io.StringIO("This is a test. And this is another one.")
    )

    res = await analyze.analyze(plugins, beanstalk_client=False)

    # Two plugins + text
    assert len(res.keys()) == 5

    assert "count" in res
    assert "sentences" in res

    assert len(res["count"]) == 2

    assert res["count"]["This is a test"] == 14
    assert res["count"]["And this is another one"] == 23



from typing import Optional

from pydantic import BaseModel, StrictInt, StrictStr

from act.scio.tlp import TLP


class ScioBaseDocument(BaseModel):
    """Document model"""

    filename: StrictStr
    uri: Optional[StrictStr]
    tlp: Optional[TLP]
    owner: Optional[StrictStr]
    store: bool = True


class Document(ScioBaseDocument):
    """Document model"""

    content: StrictStr


class LookupResponse(BaseModel):
    """Response model for document search"""

    filename: StrictStr
    content_type: StrictStr


class SubmitResponse(ScioBaseDocument):
    """Response model for document submit"""

    hexdigest: StrictStr
    count: StrictInt
    error: Optional[StrictStr]



"""Copyright 2020 mnemonic AS <opensource@mnemonic.no>

Permission to use, copy, modify, and/or distribute this software for
any purpose with or without fee is hereby granted, provided that the
above copyright notice and this permission notice appear in all
copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL
WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE
AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.

---
Elasticsearch utilities for scio
"""

from typing import Iterator, Optional, Text, Tuple

from elasticsearch import Elasticsearch


def es_client(
    host: Text,
    port: int = 9200,
    url_prefix: Optional[Text] = None,
    username: Optional[Text] = None,
    password: Optional[Text] = None,
    timeout: int = 180,
) -> Elasticsearch:
    """Elasticsearch client"""

    connection = {"host": host, "port": port, "scheme": "http"}

    if url_prefix:
        connection["url_prefx"] = url_prefix

    if username or password:
        http_auth: Optional[Tuple[Optional[Text], Optional[Text]]] = (
            username,
            password,
        )
    else:
        http_auth = None

    return Elasticsearch([connection], timeout=timeout, http_auth=http_auth)


def aggregation(
    client: Elasticsearch,
    term: Text,
    start: Text,
    end: Text,
    missing: bool = False,
    index: Text = "scio2",
) -> Iterator[Tuple[Text, int]]:
    """Aggregation"""

    response = client.search(
        index=index,
        body={
            "size": 0,
            "query": {
                "range": {
                    "Analyzed-Date": {
                        "gte": start,
                        "lte": end,
                    }
                }
            },
            "aggs": {
                term: {
                    "terms": {"field": term},
                }
            },
        },
    )

    for aggr in response["aggregations"][term]["buckets"]:
        yield aggr["key"], aggr["doc_count"]



"""
    Support for vocularies with support for
    - aliases
    - automatic generate regular expression from aliases
    - NLP stemming
"""


import configparser
import re
import sys
from logging import info
from typing import Any, Callable, Dict, List, Optional, Pattern, Text, Union

import addict
import nltk

import act.scio.aliasregex as aliasregex
from act.scio.alias import parse_aliases

DEFAULT_CONFIG = addict.Dict(
    {
        "alias": None,
        "default": None,
        "regexfromalias": False,
        "primary": False,
        "key_mod": "lower",
        "regexmanual": "",
        "object_type": None,
    }
)


def identity(x: Any) -> Any:
    """Identity function. Used as default to return same value as input"""
    return x


def from_config(config_filename: str) -> addict.Dict:
    """
    Load vocabularies from config and return Dict
    with vocabulary-name as key and Vocabulary as value
    """
    cparser = configparser.ConfigParser()
    cparser.read(str(config_filename))

    vocabularies = addict.Dict()

    for vocabulary_name in cparser.sections():
        section = cparser[vocabulary_name]

        # Config
        config = addict.Dict()
        config.alias = section.get("alias", DEFAULT_CONFIG.alias)
        config.default = section.get("default", DEFAULT_CONFIG.default)
        config.regexfromalias = section.getboolean("regexfromalias")
        config.primary = section.getboolean("primary")
        config.key_mod = section.get("key_mod", DEFAULT_CONFIG.key_mod)
        config.regexmanual = section.get("regexmanual", DEFAULT_CONFIG.regexmanual)
        config.object_type = section.get("object_type", DEFAULT_CONFIG.object_type)
        vocabularies[vocabulary_name] = Vocabulary(config)

    return vocabularies


class IllegalVocabularyKeyType(Exception):
    """Non existing Vocabulary Key Type"""

    pass


class Vocabulary:
    """
    Support for vocularies with support for
    - aliases
    - automatic generate regular expression from aliases
    - NLP stemming
    """

    def __init__(
        self, config: Union[addict.Dict, Dict[Text, Text], configparser.SectionProxy]
    ) -> None:
        """
        Args:
            config (addict.Dict):  addict.Dict with the following keys
                    * alias (str)            # Config filw with aliases
                    * primary (bool)         # Point aliases to primary name
                    * default (bool)         # Default value for if key does not exist
                    * regexfromalias (bool)  # Create regex from aliases?
                    * primary (bool)         # Use primary name as default return value
                    * key_mod (str)          # Default key_mod (default = "lower")
                    * regexmanual (str[])    # Extra regular expressions for regex search
                    * object_type (str)      # ACT object type applicable for this vocabulary
        """
        self.config = addict.Dict(DEFAULT_CONFIG)
        self.config.update(config)
        self.regex: List[Pattern[Text]] = []
        self.vocab: Dict[str, Dict[str, addict.Dict]] = addict.Dict(
            none=addict.Dict(),
            lower=addict.Dict(),
            stem=addict.Dict(),
            norm=addict.Dict(),
        )

        self.stemmer = nltk.stem.PorterStemmer().stem

        if self.config.alias:
            self.load_alias(self.config.alias)

        if self.config.regexmanual:
            self.regex = [
                re.compile(entry.strip(), re.I)
                for entry in self.config.regexmanual.split("\n")
                if entry.strip()
            ]
        else:
            self.regex = []

        if self.config.regexfromalias:
            for aliasre in aliasregex.get_reg_ex_set(self.config.alias):
                try:
                    self.regex.append(re.compile(aliasre, re.IGNORECASE))
                except re.error:
                    sys.stderr.write(f"ERROR in regex {aliasre}\n")
                    raise

    def load_alias(self, filename: str) -> None:
        """
        Load aliases from file.

        Args:
            filename (str): Filename to read aliases from
        """
        with open(filename, "r") as f:

            for line in f.readlines():
                # Remove comments (starting with '#'), unless they are escaped
                line = re.sub(r"(?<!\\)#.*", "", line)

                if not line.strip():
                    continue  # Skip empty lines

                # Get key and aliases
                primary, aliases = parse_aliases(line)

                for name in [primary] + aliases:
                    # Stor a separate entry for each type of key_mod
                    # For each entry, also store the value itself and the
                    # primary name

                    self.vocab["none"][name] = addict.Dict(value=name, primary=primary)
                    self.vocab["lower"][name.lower()] = addict.Dict(
                        value=name, primary=primary
                    )
                    self.vocab["stem"][self.stemmer(name)] = addict.Dict(
                        value=name, primary=primary
                    )

                    self.vocab["norm"][aliasregex.normalize(name)] = addict.Dict(
                        value=name, primary=primary
                    )

    def __getitem__(self, key: str) -> Optional[str]:
        """
        Return entry from vocabulary.
        """

        return self.get(key)

    def regex_search(
        self,
        text: Text,
        key_mod: Text = "DEFAULT",
        normalize_result: Callable[[Text], Text] = identity,
        debug: bool = False,
    ) -> List[str]:
        """
        Find all matches in vocabulary from regex search

        Args:
            text (str):       Input text
            key_mod (text):   Lookup using modifie key
        """

        key_mod = self.get_key_mod(key_mod)

        result = []
        for regex in self.regex:
            for match in regex.findall(text):
                if debug:
                    info("%s found by regex %s", match, regex)
                result.append(normalize_result(match))

        return result

    def get_key_mod(self, key_mod: Optional[str]) -> str:
        """Verify and get key_mod"""

        # Get key modifier from config
        if key_mod == "DEFAULT":
            key_mod = self.config.key_mod

        elif key_mod is None:  # Use "none" modifier (the value itself)
            key_mod = "none"

        if key_mod not in self.vocab:
            raise IllegalVocabularyKeyType("Illegal key: {}".format(key_mod))
        return key_mod

    def get(
        self,
        key: str,
        key_mod: str = "DEFAULT",
        primary: Optional[bool] = None,
        default: Optional[str] = None,
    ) -> Optional[str]:
        """
        Search for entry in vocabulary

        Args:
            key (str):           Key to search for
            key_mod (str):       Key modifier ("lower", "stem" or "none"). If not set,
                                 use default value form self.config.key_mod.
            primary (bool|None): If True, return primary name, if not return the
                                 value itself from the vocabulary
            default (str|None):  Default value if key is not found. Default == None
        """

        key_mod = self.get_key_mod(key_mod)

        # If primary is not set, get from config whether we should retrieve the primary name
        if primary is None:
            primary = self.config.primary

        # If default is None, get default value from config
        if default is None:
            default = self.config.default

        if key_mod == "stem":
            key = self.stemmer(key)
        elif key_mod == "lower":
            key = key.lower()
        elif key_mod == "norm":
            key = aliasregex.normalize(key)

        value = self.vocab[key_mod].get(key)

        if not value:
            return default

        if primary:  # Return primary value
            return value.primary  # type: ignore

        # Return value ifself
        return value.value  # type: ignore



"""Helper function for data extraction"""

import argparse
import html
import logging
import os.path
import urllib.parse
import uuid
from typing import Any, Dict, FrozenSet, List, Optional, Text, Tuple

import justext
import requests
from bs4 import BeautifulSoup

from act.scio.feeds import analyze, download


def get_content_from_entry(entry: Dict[Text, Any]) -> Text:
    """Extract and concatenate the content portion of a feed entry"""

    if "content" in entry:
        return "\n".join([x["value"] for x in entry["content"]])

    logging.warning("No content for %s", entry["link"])

    return ""


def get_summary_from_entry(entry: Dict[Text, Any]) -> Text:
    """Extract summary from feed entry"""

    if "summary_detail" in entry:
        value: Text = entry["summary_detail"]["value"]
        return value
    return ""


def create_html(entry: Dict[Text, Text]) -> Text:
    """Wrap an entry in html headers and footers"""

    html_data = """<html>
    <head>
        <title>{title}</title>
    </head>
    <body>
        <p>
        {summary}
        </p>
        {content}
    </body>
</html>"""

    content = get_content_from_entry(entry)
    summary = get_summary_from_entry(entry)

    return html_data.format(
        title=entry.get("title", "NO TITLE"), content=content, summary=summary
    )


def replace_unsafe_filename_characters(path: Text) -> Text:
    """Make filename safe by only allowing alpha numeric characters,
    digits and ._-"""

    def _safe_char(c: Text) -> Text:
        if c.isalpha():
            return c
        if c.isdigit():
            return c
        if c in "_-.":
            return c
        if c in " \t":
            return "_"
        return ""

    return "".join(_safe_char(c) for c in path)


def read_stoplist(fname: Text) -> FrozenSet[Text]:
    """Take a list of words (one pr. line) and create a frozenset
    for use as stopwords when extracting text with jusText"""

    with open(fname, "r", encoding="utf-8") as stream:
        return frozenset(line.strip().lower() for line in stream)


def sanitize_filename(filename: Text) -> Text:
    """make sure that the filename is usable"""

    # make sure that the basename is less than 255 characters
    basename = os.path.basename(filename)
    directory = os.path.dirname(filename)
    if len(basename) >= 255:
        filename, extension = os.path.splitext(basename)
        basename = filename[: 254 - len(extension)] + extension

    return os.path.join(directory, basename)


def create_storage_path(
    filename: Text, extension: Optional[Text], *path_elements: Text
) -> Text:
    """Build a path, sanitize filenames and build proper path structure"""

    if extension is not None:
        if not extension.startswith("."):
            extension = "." + extension
    else:
        extension = ""

    return sanitize_filename(
        os.path.join(
            *path_elements, replace_unsafe_filename_characters(filename) + extension
        )
    )


def partial_entry_text_to_file(
    args: argparse.Namespace, entry: Dict[Text, Text]
) -> Tuple[Optional[Text], Optional[Text], Optional[Text]]:
    """Download the original content and write it to the proper file.
    Return the html."""

    if "link" not in entry:
        logging.warning("entry does not contain 'link'")
        return None, None, None

    url = entry["link"]

    req = requests.get(
        url,
        proxies=download.proxies(args.proxy_string),
        headers=download.default_headers(),
        verify=False,
        timeout=60,
    )

    if req.status_code >= 400:
        logging.warning("Unable to download content: %s", url)
        return None, None, None

    filename = entry.get("title", str(uuid.uuid4()))

    html_data = "<html>\n<head>\n"
    html_data += "<title>{0}</title>\n</head>\n".format(entry.get("title", "NO TITLE"))
    html_data += "<body>\n"

    raw_html = req.text

    stoplist = (
        read_stoplist(args.stoplist)
        if args.stoplist
        else justext.get_stoplist("English")
    )

    paragraphs = justext.justext(raw_html, stoplist)

    for para in paragraphs:
        if not para.is_boilerplate:
            if para.is_heading:
                html_data += "\n<h1>{0}</h1>\n".format(html.escape(para.text))
            else:
                html_data += "<p>\n{0}\n</p>\n".format(html.escape(para.text))

    html_data += "\n</body>\n</html>"

    full_filename = create_storage_path(filename, "html", args.store_path, "download")

    with open(full_filename, "w", encoding="utf-8") as html_file:
        html_file.write(html_data)

    # we want to return the raw_html and not the "article extraction"
    # since we want to extract links to .pdfs etc.
    return full_filename, url, raw_html


def entry_text_to_file(
    args: argparse.Namespace, entry: Dict[Text, Text]
) -> Tuple[Optional[Text], Optional[Text], Optional[Text]]:
    """Extract the entry content and write it to the proper file.
    Return the wrapped HTML"""

    filename = entry.get("title", str(uuid.uuid4()))
    url = entry.get("link")

    html_data = create_html(entry)

    full_filename = create_storage_path(filename, "html", args.store_path, "download")

    with open(full_filename, "w", encoding="utf-8") as html_file:
        html_file.write(html_data)

    return full_filename, url, html_data


def get_links(
    entry: Dict[Text, Text], html_data: Text
) -> List[urllib.parse.ParseResult]:
    """Extract any links from the html"""

    links = []
    soup = BeautifulSoup(html_data, "html.parser")
    if soup:
        links = [a["href"] for a in soup.findAll("a", href=True)]
    else:
        logging.warning("soup is none : %s", entry.get("title", "NO TITLE"))

    return [analyze.parse_and_correct_link(link) for link in links]



import os

import addict
import pytest

from act.scio.plugins import threatactor_pattern


@pytest.mark.asyncio  # type: ignore
async def test_threatactor_pattern() -> None:
    """test for plugins"""

    nlpdata = addict.Dict()

    nlpdata.content = """

    Lorem Ipsum Dirty Panda, APT1 APT-2, Apt 35, APT_46, UNC2354

    The threat actors aqua blizzard was previous known as ACTINIUM

    Pumpkin Sandstorm was known as DEV-0146

    DEV-0257 is know renamed to Storm-0257.

    """

    plugin = threatactor_pattern.Plugin()
    plugin.configdir = os.path.join(
        os.path.dirname(__file__), "../act/scio/etc/plugins"
    )
    res = await plugin.analyze(nlpdata)

    assert "Dirty Panda" in res.result.ThreatActors
    assert "APT 1" in res.result.ThreatActors
    assert "APT 2" in res.result.ThreatActors
    assert "APT 35" in res.result.ThreatActors
    assert "APT 46" in res.result.ThreatActors
    assert "UNC 2354" in res.result.ThreatActors
    assert "Aqua Blizzard" in res.result.ThreatActors
    assert "Pumpkin Sandstorm" in res.result.ThreatActors
    assert "Storm-0257" in res.result.ThreatActors



import os

import addict
import pytest

from act.scio.plugins import locations


@pytest.mark.asyncio  # type: ignore
async def test_locations() -> None:
    """test for plugins"""

    nlpdata = addict.Dict()
    nlpdata.content = (
        "The Democratic Republic of Congo and the Arabic Emirates.\n\n"
        + "In London, people eat a lot of fish and chips.\n\n"
        + "England and Scotland is part of the UK.\n"
    )
    nlpdata.pos_tag = addict.Dict()
    nlpdata.pos_tag.tokens = [
        ("The", "DT"),
        ("Democratic", "JJ"),
        ("Republic", "NNP"),
        ("of", "IN"),
        ("Congo", "NNP"),
        ("and", "CC"),
        ("the", "DT"),
        ("Arabic", "NNP"),
        ("Emirates", "NNP"),
        (".", "."),
        ("In", "IN"),
        ("London", "NNP"),
        (",", ","),
        ("people", "NNS"),
        ("eat", "VBP"),
        ("a", "DT"),
        ("lot", "NN"),
        ("of", "IN"),
        ("fish", "JJ"),
        ("and", "CC"),
        ("chips", "NNS"),
        (".", "."),
        ("England", "NNP"),
        ("and", "CC"),
        ("Scotland", "NNP"),
        ("is", "VBZ"),
        ("part", "NN"),
        ("of", "IN"),
        ("the", "DT"),
        ("UK", "NNP"),
        (".", "."),
    ]

    plugin = locations.Plugin()
    plugin.configdir = os.path.join(
        os.path.dirname(__file__), "../act/scio/etc/plugins"
    )
    res = await plugin.analyze(nlpdata)

    for country in ["UK", "Republic of Congo", "England", "Scotland", "Congo"]:
        assert country in res.result.countries_mentioned

    assert "London" in [x.name for x in res.result.cities]
    assert "Congo" in [x.name for x in res.result.countries]
    assert "United Kingdom of Great Britain and Northern Ireland" in [
        x.name for x in res.result.countries_inferred
    ]



import os

import addict
import pytest

from act.scio.plugins import threatactor_nlp


@pytest.mark.asyncio  # type: ignore
async def test_sectors() -> None:
    """test for plugins"""

    nlpdata = addict.Dict()
    nlpdata.pos_tag = addict.Dict()
    nlpdata.pos_tag.tokens = [
        ["The", "DT"],
        ["Noughty", "NNP"],
        ["Noo-Noo", "NNP"],
        ["threat", "NN"],
        ["group", "NN"],
        ["sucks", "VBZ"],
        ["up", "RP"],
        ["tubby", "JJ"],
        ["toast", "NN"],
        [".", "."],
        ["The", "DT"],
        ["Tinky", "NNP"],
        ["Winky", "NNP"],
        [",", ","],
        ["Dipsy", "NNP"],
        [",", ","],
        ["Laa-Laa", "NNP"],
        ["and", "CC"],
        ["Poo", "NNP"],
        ["adversary", "JJ"],
        ["groups", "NNS"],
        ["looks", "VBZ"],
        ["sketchy", "NNS"],
        [".", "."],
    ]

    nlpdata.content = (
        "The Noughty Noo-Noo threat group sucks up tubby toast. "
        + "The Tinky Winky, Dipsy, Laa-Laa and Poo adversary groups looks sketchy."
    )

    plugin = threatactor_nlp.Plugin()
    plugin.configdir = os.path.join(
        os.path.dirname(__file__), "../act/scio/etc/plugins"
    )
    res = await plugin.analyze(nlpdata)

    tas = ["Noughty Noo-Noo", "Tinky Winky", "Dipsy", "Laa-Laa", "Poo"]

    for ta in tas:
        assert ta in res.result.actors



import configparser
import csv
import json
import os
from typing import Any, Dict, List, Set, Text, Tuple

import addict

from act.scio.plugin import BasePlugin, Result
from act.scio.vocabulary import Vocabulary


class Plugin(BasePlugin):
    name = "locations"
    info = "Extract locations from a body of text"
    version = "0.1"
    dependencies: List[Text] = ["pos_tag"]

    def nouns(self, tok: List[Tuple[Text, Text]]) -> List[Text]:
        """Rebuild a list of nouns from the tokenized values. e.g.
        [('The', 'DT'), ('Arabic', 'NNP'), ('Emirates', 'NNP')] will return
        ["Arabic Emirates"]"""

        res: Set[Text] = set()
        curr_name: List[Text] = []
        curr_in_name: List[Text] = []
        for i, (token, tag) in enumerate(tok):
            if tag == "NNP":
                # do not keep first part of Noun containing a preposition/subordinating conjunction.
                # e.g. do not keep "Republic" when what we are working on is "Republic of Congo"
                if i < len(tok) - 1 and tok[i + 1][1] != "IN":
                    curr_name.append(token)
                curr_in_name.append(token)
            elif tag == "IN":
                if curr_in_name:
                    curr_in_name.append(token)
                if curr_name:
                    res.add(" ".join(curr_name))
                    curr_name = []
            else:
                if curr_name:
                    res.add(" ".join(curr_name))
                    curr_name = []
                if curr_in_name:
                    res.add(" ".join(curr_in_name))
                    curr_in_name = []

        if curr_name:
            res.add(" ".join(curr_name))
        if curr_in_name:
            res.add(" ".join(curr_in_name))

        return list(res)

    def cities_from_file(self, filename: Text) -> Dict[Text, Dict[Text, Any]]:
        reader = csv.reader(open(filename), dialect="excel-tab")
        ret: Dict[Text, Dict[Text, Any]] = {}
        for _, city, _, _, _, _, _, _, cc, _, _, _, _, _, pop, _, _, area, _ in reader:
            city = city.split(",")[0]
            if (city in ret and ret[city]["population"] < int(pop)) or city not in ret:
                ret[city] = addict.Dict(
                    {
                        "name": city,
                        "population": int(pop),
                        "country code": cc,
                        "area": area,
                    }
                )
        return ret

    def countries_from_file(
        self, filename: Text
    ) -> Tuple[Dict[Text, Text], Dict[Text, Text]]:

        names: Dict[Text, Text] = {}
        alpha2: Dict[Text, Text] = {}

        countries = json.load(open(filename))

        for country in countries:
            names[country["name"]] = addict.Dict(country)
            alpha2[country["alpha-2"]] = addict.Dict(country)

        return names, alpha2

    async def analyze(self, nlpdata: addict.Dict) -> Result:

        res = addict.Dict()

        ini = configparser.ConfigParser()
        ini.read([os.path.join(self.configdir, "locations.ini")])
        ini["locations"]["cities"] = os.path.join(
            self.configdir, "../../vendor", ini["locations"]["cities"]
        )
        ini["locations"]["countries"] = os.path.join(
            self.configdir, "../../vendor", ini["locations"]["countries"]
        )
        ini["vocabulary"]["alias"] = os.path.join(
            self.configdir, ini["vocabulary"]["alias"]
        )

        cities = self.cities_from_file(ini["locations"]["cities"])
        country_names, country_cc = self.countries_from_file(
            ini["locations"]["countries"]
        )

        nouns = self.nouns(nlpdata.pos_tag.tokens)
        vocab = Vocabulary(ini["vocabulary"])

        res.cities = []
        res.countries = []
        res.countries_inferred = []
        res.countries_mentioned = []

        for noun in nouns:
            if noun in cities:
                city = cities[noun]
                res.cities.append(city)
                res.countries_inferred.append(
                    country_cc.get(city["country code"], "UNK")
                )
            if noun in country_names:
                res.countries.append(country_names[noun])
            if vocab.get(noun):
                res.countries_mentioned.append(noun)

        return Result(name=self.name, version=self.version, result=res)



"Alias regex tests"
from act.scio.aliasregex import normalize


def test_aliasregex_normalization() -> None:
    "Normalize test"

    assert normalize("oceanlotusGroup") == "oceanlotus group"
    assert normalize("APT 27") == "apt 27"
    assert normalize("APT.27") == "apt 27"
    assert normalize("APT-27") == "apt 27"
    assert normalize("APT- 27") == "apt 27"
    assert normalize("winntiGroup") == "winnti group"



#!/usr/bin/env python3

import argparse
import base64
import hashlib
import os.path
import sys
from typing import Any, Dict, Optional, Text, Tuple

import requests

from act.scio import tlp
from act.scio.models import Document

"Submit content of URI / File to SCIO"


def fatal(message: Text, exit_code: int = 1) -> None:
    "Print error message and abort"
    sys.stderr.write(message + "\n")
    sys.exit(exit_code)


VALID_TLP = ", ".join(tlp.TLP_ALLOWED_VALUES)


DEFAULT_USER_AGENT = (
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) "
    + "Chrome/112.0.0.0 Safari/537.36"
)


def parseargs(description: Text) -> argparse.Namespace:
    """Parse arguments"""
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("--http-timeout", type=int, default=120, help="Timeout")
    parser.add_argument("--proxy-string", help="Proxy to use for external queries")
    parser.add_argument("--scio-baseuri", required=True, help="SCIO API URI")
    parser.add_argument("--http-user", help="SCIO HTTP Basic Auth user")
    parser.add_argument(
        "--http-user-agent",
        default=DEFAULT_USER_AGENT,
        help="User agent used to download from uri",
    )
    parser.add_argument("--http-password", help="SCIO HTTP Basic Auth password")
    parser.add_argument("--uri", help="Download content from URI and submit content")
    parser.add_argument("--filename", help="Submit file")
    parser.add_argument(
        "--nostore",
        action="store_true",
        help="Do not store document to elasticsearch/disk",
    )
    parser.add_argument("--tlp", default="AMBER", help=f"tlp ({VALID_TLP})")
    parser.add_argument("--owner", help="Document owner (identifier)")
    args = parser.parse_args()

    if not (args.uri or args.filename):
        fatal("Specify either --uri or --filename")

    args.tlp = args.tlp.upper()

    if not tlp.valid_tlp(args.tlp):
        fatal(f"Invalid TLP: {args.tlp}. Must be one of: {VALID_TLP}")

    return args


def get_uri(
    uri: Text,
    proxies: Optional[Dict[Text, Any]],
    timeout: Optional[int] = 30,
    user_agent: Optional[str] = None,
) -> bytes:
    "Return content from URI"

    headers = {"User-Agent": user_agent} if user_agent else {}

    req = requests.get(uri, proxies=proxies, timeout=timeout, headers=headers)

    if not req.status_code == 200:
        fatal("Unable to get {}. Status code={}".format(uri, req.status_code))

    return req.content


def get_file(filename: Text) -> bytes:
    "Return content from file"
    with open(filename, "rb") as f:
        return f.read()


def scio_submit(
    submit_uri: Text,
    content: bytes,
    filename: Text,
    owner: Optional[Text],
    auth: Optional[Tuple[Text, Text]],
    uri: Optional[Text],
    tlp: tlp.TLP = "AMBER",
    store: bool = False,
) -> None:
    "Submit content with specified filename to SCIO"

    sha256 = hashlib.sha256(content).hexdigest()

    # If filename is not specified, use sha256
    if not filename:
        filename = sha256

    document = Document(
        content=base64.b64encode(content).decode("utf8"),
        tlp=tlp,
        filename=filename,
        owner=owner,
        store=store,
        uri=uri,
    )

    req = requests.post(submit_uri, json=document.dict(), auth=auth)

    if not req.status_code == 200:
        fatal(f"Unable to submit {filename} to scio ({submit_uri}), error={req.text}")

    print(f"Submitted {sha256}: {req.text}")


def main() -> None:
    "main function"

    args = parseargs("SCIO submit from URI/File")
    auth = (args.http_user, args.http_password) if args.http_user else None

    proxies = (
        {"http": args.proxy_string, "https": args.proxy_string}
        if args.proxy_string
        else None
    )

    if args.uri:
        content = get_uri(
            args.uri,
            proxies=proxies,
            timeout=args.http_timeout,
            user_agent=args.http_user_agent,
        )
        # Use basename from URI as filename
        filename = os.path.basename(args.uri)

    elif args.filename:
        content = get_file(args.filename)
        # Use basename from file
        filename = os.path.basename(args.filename)

    try:
        scio_submit(
            args.scio_baseuri,
            content,
            filename,
            args.owner,
            auth,
            args.uri if args.uri else None,
            args.tlp,
            not args.nostore,
        )
    except requests.exceptions.ConnectionError as e:
        fatal(f"Unable to connect to {args.scio_baseuri}: {e}")


if __name__ == "__main__":
    main()



""" test feed analyse """

import urllib.parse

from act.scio.feeds import analyze


def test_filter_links() -> None:
    """test that links filter works"""

    links = [
        "http://example.com/check.js",
        "https://example.com/some.pdf",
        "http://example.com/sitemap.xml",
    ]
    correct = ["https://example.com/some.pdf", "http://example.com/sitemap.xml"]

    parsed_links = [urllib.parse.urlparse(link) for link in links]
    parsed_correct = [urllib.parse.urlparse(link) for link in correct]

    assert parsed_correct == analyze.filter_links(["xml", "pdf"], parsed_links)


def test_filter_links_with_exlude() -> None:
    """test that links filter works"""

    links = [
        "http://example.com/check.js",
        "https://example.com/some.pdf",
        "http://example.com/sitemap.xml",
    ]
    correct = ["https://example.com/some.pdf"]

    parsed_links = [urllib.parse.urlparse(link) for link in links]
    parsed_correct = [urllib.parse.urlparse(link) for link in correct]

    assert parsed_correct == analyze.filter_links(
        ["xml", "pdf"], parsed_links, exclude_filenames=["sitemap.xml"]
    )



import os

import addict
import pytest

from act.scio.plugins import sectors


@pytest.mark.asyncio  # type: ignore
async def test_sectors() -> None:
    """test for plugins"""

    nlpdata = addict.Dict()
    nlpdata.pos_tag = addict.Dict()
    nlpdata.pos_tag.tokens = [
        ("The", "DT"),
        ("companies", "NNS"),
        ("in", "IN"),
        ("the", "DT"),
        ("Bus", "NNP"),
        (";", ":"),
        ("Finanical", "NNP"),
        (",", ","),
        ("Aviation", "NNP"),
        ("and", "CC"),
        ("Automobile", "NNP"),
        ("industry", "NN"),
        ("is", "VBZ"),
        ("in", "IN"),
        ("trouble", "NN"),
        (".", "."),
    ]

    nlpdata.content = (
        "The companies in the Bus; Finanical, "
        + "Aviation and Automobile industry are large."
    )

    plugin = sectors.Plugin()
    plugin.configdir = os.path.join(
        os.path.dirname(__file__), "../act/scio/etc/plugins"
    )
    res = await plugin.analyze(nlpdata)

    assert "aerospace" in res.result.sectors
    assert "automotive" in res.result.sectors
    assert "finanical" not in res.result.sectors
    assert "finanical" not in res.result.unknown_sectors



import os

import addict
import pytest

from act.scio.plugins import mitre_attack

TEST_TEXT = """
Technical Analysis of Operation Diànxùn18REPORTTa c t i c
TechniqueObservableIOCsResource DevelopmentAcquire Infrastructure: Domains
(T1583.001)Attackers purchased domains to develop their phishing attack.
“flach.cn”“careerhuawei.net”Develop capabilities: Malware (T1587.001)Attackers
built malicious components to conduct their attack. Fake FlashUtility
DownloaderAddTaskPlanDllVersion.dllObtain capabilities: Tool
(T1588.002)Attackers acquired red teaming tools to conduct their attack.Cobalt
Strike Initial AccessSpear phishing Link (T1566.002)ExecutionUser Execution
(T1204.001)Users are redirected to the Fake Flash website to download the first
stage.Persistence Scheduled Task/Job: Scheduled Task (T1053.005)The DotNet
Utility creates a scheduled task that will run cmd.exe /c with the previous
payload downloaded and create registry key.“SOFTWARE\\Microsoft\\Windows
NT\\CurrentVersion\\AppCompatFlags\\TelemetryController\\Levint”Create or
Modify System Process: Windows Service (T1543.003)The DotNet utility can add a
WMI backdoor by creating a permanent filter in order to stay persistent in the
infected machine.Defense EvasionProcess Injection (T1055)The DotNet utility has
the possibility to inject shellcode into the clipboard.Command And Control
Application Layer Protocol: Web Proto

G1234
G4321.
S1234,S4321.
TA1234.
TA4321
"""


@pytest.mark.asyncio  # type: ignore
async def test_sectors() -> None:
    """test for plugins"""

    nlpdata = addict.Dict()
    plugin = mitre_attack.Plugin()
    plugin.configdir = os.path.join(
        os.path.dirname(__file__), "../act/scio/etc/plugins"
    )

    nlpdata.content = TEST_TEXT

    res = await plugin.analyze(nlpdata)

    assert len(res.result.Groups) == 2
    assert "G4321" in res.result.Groups
    assert "G1234" in res.result.Groups
    assert len(res.result.Tactics) == 2
    assert "TA4321" in res.result.Tactics
    assert "TA1234" in res.result.Tactics
    assert "T1055" in res.result.Techniques
    assert "T1055" not in res.result.SubTechniques
    assert len(res.result.SubTechniques) == 7
    assert "T1583.001" in res.result.SubTechniques
    assert "T1587.001" in res.result.SubTechniques
    assert "T1588.002" in res.result.SubTechniques
    assert "T1566.002" in res.result.SubTechniques
    assert "T1204.001" in res.result.SubTechniques
    assert "T1053.005" in res.result.SubTechniques
    assert len(res.result.Software) == 2
    assert "S1234" in res.result.Software
    assert "S4321" in res.result.Software



"""Tika/Scio integration. This module contains the scio tike server and service code"""

import argparse
import asyncio
import gzip
import html
import json
import logging
import os
import time
from typing import Any, Optional, Text

import caep
import greenstalk
import tika
from tika import parser

import act.scio.config
import act.scio.logsetup


def parse_args() -> argparse.Namespace:
    """Helper setting up the argsparse configuration"""

    arg_parser = act.scio.config.parse_args("Scio 2 Tika server")
    return caep.config.handle_args(arg_parser, "scio/etc", "scio.ini", "tika")  # type: ignore


class Server:
    """The server class listening for new work on beanstalk and sending it to
    Apache Tika for text extraction and den sending it to Scio for text analyzis."""

    def __init__(self, beanstalk_host: Text = "127.0.0.1", beanstalk_port: int = 11300):
        self.client: Optional[greenstalk.Client] = None
        self.connect(beanstalk_host, beanstalk_port)

        logging.info("initialize tika VM")
        tika.initVM()

    def client_ready(self) -> bool:
        """client_ready is a utility function checking whether the
        beanstalk client has a connection that works."""

        if not self.client:
            return False
        try:
            self.client.stats()
            return True
        except (OSError, BrokenPipeError, ConnectionError, ConnectionRefusedError):
            return False

    def connect(
        self, beanstalk_host: Text = "127.0.0.1", beanstalk_port: int = 11300
    ) -> None:
        """Try to create a beanstalk connection"""

        if self.client:
            # close existing connection
            self.client.close()

        while not self.client_ready():
            try:
                logging.info(
                    "Trying to connect to beanstalkd on %s:%s",
                    beanstalk_host,
                    beanstalk_port,
                )
                self.client = greenstalk.Client((beanstalk_host, beanstalk_port))
            except ConnectionRefusedError as err:
                logging.warning("Server.connect: %s", err)
                logging.info("Server.connect: waiting 5 seconds")
                time.sleep(5)

        # We only want to receive messages specifically to scio. The default
        # beanstalk tube is ignored.
        self.client.watch("scio_doc")
        self.client.ignore("default")
        self.client.use("scio_analyze")

    async def reserve(self) -> Any:
        """Async reserve"""
        return self.client.reserve()  # type: ignore

    async def worker(self, worker_id: int) -> None:
        """Main worker code. Listening to the beanstalk client for ready work, sending it to the
        tika service and then posting the extracted document to the queue for consumption by
        the analyzis module"""

        while True:
            logging.info("Worker [%s] waiting for work.", worker_id)
            try:
                job = await self.reserve()
            except greenstalk.TimedOutError:
                continue

            logging.info("Worker [%s] got job.", worker_id)
            try:
                meta_data = json.loads(job.body)
            except json.decoder.JSONDecodeError as err:
                logging.warning("Unable to decode body %s [%s ...]", err, job.body[:25])
                self.client.delete(job)  # type: ignore
                continue

            if "filename" not in meta_data:
                logging.warning("No 'filename' field in meta data : %s", meta_data)
                self.client.delete(job)  # type: ignore
                continue

            if not os.path.isfile(meta_data["filename"]):
                logging.warning("Could not find file '%s'", meta_data["filename"])
                self.client.delete(job)  # type: ignore
                continue

            with open(meta_data["filename"], "rb") as fh:
                content = fh.read()
                if meta_data["filename"].endswith(".html"):
                    content = html.unescape(content.decode("utf8")).encode("utf8")
                data = parser.from_buffer(content)
                data.update(meta_data)

            self.client.delete(job)  # type: ignore
            logging.info("Worker [%s] waiting to post result.", worker_id)
            try:
                self.client.put(  # type: ignore
                    gzip.compress(json.dumps({**data, **meta_data}).encode("utf8"))
                )
                logging.info("Worker [%s] job done.", worker_id)
            except greenstalk.JobTooBigError:
                logging.error("Job to big: %s.", meta_data["filename"])

    async def _start(self, n: int) -> None:
        """Start the server, create n number of workers and wait for data on the queue"""

        loop = asyncio.get_event_loop()
        workers = []

        for i in range(n):
            logging.info("Starting worker [%s]", i)
            workers.append((n, loop.create_task(self.worker(i))))

        logging.info("Gather all workers")
        asyncio.gather(*[w for (_, w) in workers])

        for i, worker in workers:
            if worker.exception():
                logging.error(
                    "Worker [%s] ended with exception %s", i, worker.exception()
                )

    def start(self, n: int = 4) -> None:
        """start the server"""

        loop = asyncio.get_event_loop()
        loop.run_until_complete(asyncio.wait([self._start(n)]))

        logging.info("Server ended")


def main() -> None:
    """entry point"""

    args = parse_args()

    act.scio.logsetup.setup_logging(args.loglevel, args.logfile, "scio-tika-server")

    server = Server(args.beanstalk, args.beanstalk_port)

    logging.info("Starting Tika server")
    server.start()

    logging.info("Finnished Tika server")


if __name__ == "__main__":
    main()
